@@ -24,25 +24,30 @@
 // IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 // DEALINGS IN THE SOFTWARE.
 
-use crate::{response, AccessControl, TEN_MB_SIZE_BYTES};
+use crate::{response, AccessControl};
 use futures_channel::mpsc;
-use futures_util::stream::StreamExt;
+use futures_util::{lock::Mutex, stream::StreamExt, SinkExt};
 use hyper::{
 	server::{conn::AddrIncoming, Builder as HyperBuilder},
 	service::{make_service_fn, service_fn},
 	Error as HyperError,
 };
-use jsonrpsee_types::error::{Error, GenericTransportError};
-use jsonrpsee_types::v2::error::JsonRpcErrorCode;
-use jsonrpsee_types::v2::params::Id;
-use jsonrpsee_types::v2::request::{JsonRpcInvalidRequest, JsonRpcNotification, JsonRpcRequest};
+use jsonrpsee_types::{
+	error::{Error, GenericTransportError},
+	v2::{
+		error::JsonRpcErrorCode,
+		params::Id,
+		request::{JsonRpcInvalidRequest, JsonRpcNotification, JsonRpcRequest},
+	},
+	TEN_MB_SIZE_BYTES,
+};
 use jsonrpsee_utils::hyper_helpers::read_response_to_body;
-use jsonrpsee_utils::server::rpc_module::RpcModule;
 use jsonrpsee_utils::server::{
 	helpers::{collect_batch_response, send_error},
 	rpc_module::Methods,
 };
 
+use serde_json::value::RawValue;
 use socket2::{Domain, Socket, Type};
 use std::{
 	cmp,
@@ -95,12 +100,16 @@ impl Builder {
 		let local_addr = listener.local_addr().ok();
 
 		let listener = hyper::Server::from_tcp(listener)?;
+
+		let stop_pair = mpsc::channel(1);
 		Ok(Server {
 			listener,
 			local_addr,
 			methods: Methods::default(),
 			access_control: self.access_control,
 			max_request_body_size: self.max_request_body_size,
+			stop_pair,
+			stop_handle: Arc::new(Mutex::new(())),
 		})
 	}
 }
@@ -111,6 +120,25 @@ impl Default for Builder {
 	}
 }
 
+/// Handle used to stop the running server.
+#[derive(Debug, Clone)]
+pub struct StopHandle {
+	stop_sender: mpsc::Sender<()>,
+	stop_handle: Arc<Mutex<()>>,
+}
+
+impl StopHandle {
+	/// Requests server to stop. Returns an error if server was already stopped.
+	pub async fn stop(&mut self) -> Result<(), Error> {
+		self.stop_sender.send(()).await.map_err(|_| Error::AlreadyStopped)
+	}
+
+	/// Blocks indefinitely until the server is stopped.
+	pub async fn wait_for_stop(&self) {
+		self.stop_handle.lock().await;
+	}
+}
+
 /// An HTTP JSON RPC server.
 #[derive(Debug)]
 pub struct Server {
@@ -124,32 +152,33 @@ pub struct Server {
 	max_request_body_size: u32,
 	/// Access control
 	access_control: AccessControl,
+	/// Pair of channels to stop the server.
+	stop_pair: (mpsc::Sender<()>, mpsc::Receiver<()>),
+	/// Stop handle that indicates whether server has been stopped.
+	stop_handle: Arc<Mutex<()>>,
 }
 
 impl Server {
-	/// Register all methods from a [`Methods`] of provided [`RpcModule`] on this server.
-	/// In case a method already is registered with the same name, no method is added and a [`Error::MethodAlreadyRegistered`]
-	/// is returned. Note that the [`RpcModule`] is consumed after this call.
-	pub fn register_module<Context: Send + Sync + 'static>(&mut self, module: RpcModule<Context>) -> Result<(), Error> {
-		self.methods.merge(module.into_methods())?;
-		Ok(())
-	}
-
-	/// Returns a `Vec` with all the method names registered on this server.
-	pub fn method_names(&self) -> Vec<&'static str> {
-		self.methods.method_names()
-	}
-
 	/// Returns socket address to which the server is bound.
 	pub fn local_addr(&self) -> Result<SocketAddr, Error> {
 		self.local_addr.ok_or_else(|| Error::Custom("Local address not found".into()))
 	}
 
+	/// Returns the handle to stop the running server.
+	pub fn stop_handle(&self) -> StopHandle {
+		StopHandle { stop_sender: self.stop_pair.0.clone(), stop_handle: self.stop_handle.clone() }
+	}
+
 	/// Start the server.
-	pub async fn start(self) -> Result<(), Error> {
-		let methods = Arc::new(self.methods);
+	pub async fn start(self, methods: impl Into<Methods>) -> Result<(), Error> {
+		// Lock the stop mutex so existing stop handles can wait for server to stop.
+		// It will be unlocked once this function returns.
+		let _stop_handle = self.stop_handle.lock().await;
+
 		let max_request_body_size = self.max_request_body_size;
 		let access_control = self.access_control;
+		let mut stop_receiver = self.stop_pair.1;
+		let methods = methods.into();
 
 		let make_service = make_service_fn(move |_| {
 			let methods = methods.clone();
@@ -196,7 +225,8 @@ impl Server {
 						if let Ok(req) = serde_json::from_slice::<JsonRpcRequest>(&body) {
 							// NOTE: we don't need to track connection id on HTTP, so using hardcoded 0 here.
 							methods.execute(&tx, req, 0).await;
-						} else if let Ok(_req) = serde_json::from_slice::<JsonRpcNotification>(&body) {
+						} else if let Ok(_req) = serde_json::from_slice::<JsonRpcNotification<Option<&RawValue>>>(&body)
+						{
 							return Ok::<_, HyperError>(response::ok_response("".into()));
 						} else if let Ok(batch) = serde_json::from_slice::<Vec<JsonRpcRequest>>(&body) {
 							if !batch.is_empty() {
@@ -207,7 +237,9 @@ impl Server {
 							} else {
 								send_error(Id::Null, &tx, JsonRpcErrorCode::InvalidRequest.into());
 							}
-						} else if let Ok(_batch) = serde_json::from_slice::<Vec<JsonRpcNotification>>(&body) {
+						} else if let Ok(_batch) =
+							serde_json::from_slice::<Vec<JsonRpcNotification<Option<&RawValue>>>>(&body)
+						{
 							return Ok::<_, HyperError>(response::ok_response("".into()));
 						} else {
 							log::error!(
@@ -236,7 +268,12 @@ impl Server {
 		});
 
 		let server = self.listener.serve(make_service);
-		server.await.map_err(Into::into)
+		server
+			.with_graceful_shutdown(async move {
+				stop_receiver.next().await;
+			})
+			.await
+			.map_err(Into::into)
 	}
 }
 
